@@ -5,6 +5,7 @@ package core
 
 import (
 	"bufio"
+	"crypto/rand"
 	"crypto/sha256"
 	"encoding/json"
 	"fmt"
@@ -23,9 +24,11 @@ import (
 	"sync"
 	"syscall"
 
+	"github.com/bazelbuild/bazelisk/config"
 	"github.com/bazelbuild/bazelisk/httputil"
 	"github.com/bazelbuild/bazelisk/platforms"
 	"github.com/bazelbuild/bazelisk/versions"
+	"github.com/bazelbuild/bazelisk/ws"
 	"github.com/mitchellh/go-homedir"
 )
 
@@ -33,7 +36,6 @@ const (
 	bazelReal      = "BAZEL_REAL"
 	skipWrapperEnv = "BAZELISK_SKIP_WRAPPER"
 	wrapperPath    = "./tools/bazel"
-	rcFileName     = ".bazeliskrc"
 	maxDirLength   = 255
 )
 
@@ -49,6 +51,29 @@ var (
 // Bazel with.
 type ArgsFunc func(resolvedBazelVersion string) []string
 
+func MakeDefaultConfig() config.Config {
+	configs := []config.Config{config.FromEnv()}
+
+	workspaceConfigPath, err := config.LocateWorkspaceConfigFile()
+	if err == nil {
+		c, err := config.FromFile(workspaceConfigPath)
+		if err != nil {
+			log.Fatal(err)
+		}
+		configs = append(configs, c)
+	}
+
+	userConfigPath, err := config.LocateUserConfigFile()
+	if err == nil {
+		c, err := config.FromFile(userConfigPath)
+		if err != nil {
+			log.Fatal(err)
+		}
+		configs = append(configs, c)
+	}
+	return config.Layered(configs...)
+}
+
 // RunBazelisk runs the main Bazelisk logic for the given arguments and Bazel repositories.
 func RunBazelisk(args []string, repos *Repositories) (int, error) {
 	return RunBazeliskWithArgsFunc(func(_ string) []string { return args }, repos)
@@ -57,9 +82,16 @@ func RunBazelisk(args []string, repos *Repositories) (int, error) {
 // RunBazeliskWithArgsFunc runs the main Bazelisk logic for the given ArgsFunc and Bazel
 // repositories.
 func RunBazeliskWithArgsFunc(argsFunc ArgsFunc, repos *Repositories) (int, error) {
-	httputil.UserAgent = getUserAgent()
 
-	bazeliskHome := GetEnvOrConfig("BAZELISK_HOME")
+	return RunBazeliskWithArgsFuncAndConfig(argsFunc, repos, MakeDefaultConfig())
+}
+
+// RunBazeliskWithArgsFuncAndConfig runs the main Bazelisk logic for the given ArgsFunc and Bazel
+// repositories and config.
+func RunBazeliskWithArgsFuncAndConfig(argsFunc ArgsFunc, repos *Repositories, config config.Config) (int, error) {
+	httputil.UserAgent = getUserAgent(config)
+
+	bazeliskHome := config.Get("BAZELISK_HOME")
 	if len(bazeliskHome) == 0 {
 		userCacheDir, err := os.UserCacheDir()
 		if err != nil {
@@ -74,7 +106,7 @@ func RunBazeliskWithArgsFunc(argsFunc ArgsFunc, repos *Repositories) (int, error
 		return -1, fmt.Errorf("could not create directory %s: %v", bazeliskHome, err)
 	}
 
-	bazelVersionString, err := getBazelVersion()
+	bazelVersionString, err := getBazelVersion(config)
 	if err != nil {
 		return -1, fmt.Errorf("could not get Bazel version: %v", err)
 	}
@@ -91,7 +123,7 @@ func RunBazeliskWithArgsFunc(argsFunc ArgsFunc, repos *Repositories) (int, error
 	// If we aren't using a local Bazel binary, we'll have to parse the version string and
 	// download the version that the user wants.
 	if !filepath.IsAbs(bazelPath) {
-		bazelPath, err = downloadBazel(bazelVersionString, bazeliskHome, repos)
+		bazelPath, err = downloadBazel(bazelVersionString, bazeliskHome, repos, config)
 		if err != nil {
 			return -1, fmt.Errorf("could not download Bazel: %v", err)
 		}
@@ -108,7 +140,7 @@ func RunBazeliskWithArgsFunc(argsFunc ArgsFunc, repos *Repositories) (int, error
 	// --print_env must be the first argument.
 	if len(args) > 0 && args[0] == "--print_env" {
 		// print environment variables for sub-processes
-		cmd := makeBazelCmd(bazelPath, args, nil)
+		cmd := makeBazelCmd(bazelPath, args, nil, config)
 		for _, val := range cmd.Env {
 			fmt.Println(val)
 		}
@@ -121,12 +153,12 @@ func RunBazeliskWithArgsFunc(argsFunc ArgsFunc, repos *Repositories) (int, error
 		if err != nil {
 			return -1, err
 		}
-		newFlags, err := getIncompatibleFlags(bazelPath, cmd)
+		newFlags, err := getIncompatibleFlags(bazelPath, cmd, config)
 		if err != nil {
 			return -1, fmt.Errorf("could not get the list of incompatible flags: %v", err)
 		}
 		if args[0] == "--migrate" {
-			migrate(bazelPath, args[1:], newFlags)
+			migrate(bazelPath, args[1:], newFlags, config)
 		} else {
 			// When --strict is present, it expands to the list of --incompatible_ flags
 			// that should be enabled for the given Bazel version.
@@ -140,7 +172,7 @@ func RunBazeliskWithArgsFunc(argsFunc ArgsFunc, repos *Repositories) (int, error
 		value := args[0][len("--bisect="):]
 		commits := strings.Split(value, "..")
 		if len(commits) == 2 {
-			bisect(commits[0], commits[1], args[1:], bazeliskHome, repos)
+			bisect(commits[0], commits[1], args[1:], bazeliskHome, repos, config)
 		} else {
 			return -1, fmt.Errorf("Error: Invalid format for --bisect. Expected format: '--bisect=<good bazel commit>..<bad bazel commit>'")
 		}
@@ -166,7 +198,7 @@ func RunBazeliskWithArgsFunc(argsFunc ArgsFunc, repos *Repositories) (int, error
 		}
 	}
 
-	exitCode, err := runBazel(bazelPath, args, nil)
+	exitCode, err := runBazel(bazelPath, args, nil, config)
 	if err != nil {
 		return -1, fmt.Errorf("could not run Bazel: %v", err)
 	}
@@ -182,130 +214,14 @@ func getBazelCommand(args []string) (string, error) {
 	return "", fmt.Errorf("could not find a valid Bazel command in %q. Please run `bazel help` if you need help on how to use Bazel", strings.Join(args, " "))
 }
 
-func getUserAgent() string {
-	agent := GetEnvOrConfig("BAZELISK_USER_AGENT")
+func getUserAgent(config config.Config) string {
+	agent := config.Get("BAZELISK_USER_AGENT")
 	if len(agent) > 0 {
 		return agent
 	}
 	return fmt.Sprintf("Bazelisk/%s", BazeliskVersion)
 }
 
-// GetEnvOrConfig reads a configuration value from the environment, but fall back to reading it from .bazeliskrc in the workspace root.
-func GetEnvOrConfig(name string) string {
-	if val := os.Getenv(name); val != "" {
-		return val
-	}
-
-	fileConfigOnce.Do(loadFileConfig)
-
-	return fileConfig[name]
-}
-
-// loadFileConfig locates available .bazeliskrc configuration files, parses them with a precedence order preference,
-// and updates a global configuration map with their contents. This routine should be executed exactly once.
-func loadFileConfig() {
-	var rcFilePaths []string
-
-	if userRC, err := locateUserConfigFile(); err == nil {
-		rcFilePaths = append(rcFilePaths, userRC)
-	}
-	if workspaceRC, err := locateWorkspaceConfigFile(); err == nil {
-		rcFilePaths = append(rcFilePaths, workspaceRC)
-	}
-
-	fileConfig = make(map[string]string)
-	for _, rcPath := range rcFilePaths {
-		config, err := parseFileConfig(rcPath)
-		if err != nil {
-			log.Fatal(err)
-		}
-
-		for key, value := range config {
-			fileConfig[key] = value
-		}
-	}
-}
-
-// locateWorkspaceConfigFile locates a .bazeliskrc file in the current workspace root.
-func locateWorkspaceConfigFile() (string, error) {
-	workingDirectory, err := os.Getwd()
-	if err != nil {
-		return "", err
-	}
-	workspaceRoot := findWorkspaceRoot(workingDirectory)
-	if workspaceRoot == "" {
-		return "", err
-	}
-	return filepath.Join(workspaceRoot, rcFileName), nil
-}
-
-// locateUserConfigFile locates a .bazeliskrc file in the user's home directory.
-func locateUserConfigFile() (string, error) {
-	home, err := os.UserHomeDir()
-	if err != nil {
-		return "", err
-	}
-	return filepath.Join(home, rcFileName), nil
-}
-
-// parseFileConfig parses a .bazeliskrc file as a map of key-value configuration values.
-func parseFileConfig(rcFilePath string) (map[string]string, error) {
-	config := make(map[string]string)
-
-	contents, err := ioutil.ReadFile(rcFilePath)
-	if err != nil {
-		if os.IsNotExist(err) {
-			// Non-critical error.
-			return config, nil
-		}
-		return nil, err
-	}
-
-	for _, line := range strings.Split(string(contents), "\n") {
-		if strings.HasPrefix(line, "#") {
-			// comments
-			continue
-		}
-		parts := strings.SplitN(line, "=", 2)
-		if len(parts) < 2 {
-			continue
-		}
-		key := strings.TrimSpace(parts[0])
-		config[key] = strings.TrimSpace(parts[1])
-	}
-
-	return config, nil
-}
-
-// isValidWorkspace returns true iff the supplied path is the workspace root, defined by the presence of
-// a file named WORKSPACE or WORKSPACE.bazel
-// see https://github.com/bazelbuild/bazel/blob/8346ea4cfdd9fbd170d51a528fee26f912dad2d5/src/main/cpp/workspace_layout.cc#L37
-func isValidWorkspace(path string) bool {
-	info, err := os.Stat(path)
-	if err != nil {
-		return false
-	}
-
-	return !info.IsDir()
-}
-
-func findWorkspaceRoot(root string) string {
-	if isValidWorkspace(filepath.Join(root, "WORKSPACE")) {
-		return root
-	}
-
-	if isValidWorkspace(filepath.Join(root, "WORKSPACE.bazel")) {
-		return root
-	}
-
-	parentDirectory := filepath.Dir(root)
-	if parentDirectory == root {
-		return ""
-	}
-
-	return findWorkspaceRoot(parentDirectory)
-}
-
 // TODO(go 1.18): remove backport of strings.Cut
 func cutString(s, sep string) (before, after string, found bool) {
 	if i := strings.Index(s, sep); i >= 0 {
@@ -314,7 +230,7 @@ func cutString(s, sep string) (before, after string, found bool) {
 	return s, "", false
 }
 
-func getBazelVersion() (string, error) {
+func getBazelVersion(config config.Config) (string, error) {
 	// Check in this order:
 	// - env var "USE_BAZEL_VERSION" is set to a specific version.
 	// - workspace_root/.bazeliskrc exists -> read contents, in contents:
@@ -330,7 +246,7 @@ func getBazelVersion() (string, error) {
 	// - workspace_root/.bazeliskrc exists -> read contents, in contents:
 	//   var "USE_BAZEL_FALLBACK_VERSION" is set to a fallback version format.
 	// - fallback version format "silent:latest"
-	bazelVersion := GetEnvOrConfig("USE_BAZEL_VERSION")
+	bazelVersion := config.Get("USE_BAZEL_VERSION")
 	if len(bazelVersion) != 0 {
 		return bazelVersion, nil
 	}
@@ -340,7 +256,7 @@ func getBazelVersion() (string, error) {
 		return "", fmt.Errorf("could not get working directory: %v", err)
 	}
 
-	workspaceRoot := findWorkspaceRoot(workingDirectory)
+	workspaceRoot := ws.FindWorkspaceRoot(workingDirectory)
 	if len(workspaceRoot) != 0 {
 		bazelVersionPath := filepath.Join(workspaceRoot, ".bazelversion")
 		if _, err := os.Stat(bazelVersionPath); err == nil {
@@ -363,7 +279,7 @@ func getBazelVersion() (string, error) {
 		}
 	}
 
-	fallbackVersionFormat := GetEnvOrConfig("USE_BAZEL_FALLBACK_VERSION")
+	fallbackVersionFormat := config.Get("USE_BAZEL_FALLBACK_VERSION")
 	fallbackVersionMode, fallbackVersion, hasFallbackVersionMode := cutString(fallbackVersionFormat, ":")
 	if !hasFallbackVersionMode {
 		fallbackVersionMode, fallbackVersion, hasFallbackVersionMode = "silent", fallbackVersionMode, true
@@ -400,7 +316,7 @@ func parseBazelForkAndVersion(bazelForkAndVersion string) (string, string, error
 	return bazelFork, bazelVersion, nil
 }
 
-func downloadBazel(bazelVersionString string, bazeliskHome string, repos *Repositories) (string, error) {
+func downloadBazel(bazelVersionString string, bazeliskHome string, repos *Repositories, config config.Config) (string, error) {
 	bazelFork, bazelVersion, err := parseBazelForkAndVersion(bazelVersionString)
 	if err != nil {
 		return "", fmt.Errorf("could not parse Bazel fork and version: %v", err)
@@ -411,78 +327,129 @@ func downloadBazel(bazelVersionString string, bazeliskHome string, repos *Reposi
 		return "", fmt.Errorf("could not resolve the version '%s' to an actual version number: %v", bazelVersion, err)
 	}
 
-	bazelForkOrURL := dirForURL(GetEnvOrConfig(BaseURLEnv))
+	bazelForkOrURL := dirForURL(config.Get(BaseURLEnv))
 	if len(bazelForkOrURL) == 0 {
 		bazelForkOrURL = bazelFork
 	}
 
-	baseDirectory := filepath.Join(bazeliskHome, "downloads", bazelForkOrURL)
-	bazelPath, err := downloadBazelIfNecessary(resolvedBazelVersion, baseDirectory, repos, downloader)
+	bazelPath, err := downloadBazelIfNecessary(resolvedBazelVersion, bazeliskHome, bazelForkOrURL, repos, config, downloader)
 	return bazelPath, err
 }
 
-func downloadBazelIfNecessary(version string, baseDirectory string, repos *Repositories, downloader DownloadFunc) (string, error) {
+// downloadBazelIfNecessary returns a path to a bazel which can be run, which may have been cached.
+// The directory it returns may depend on version and bazeliskHome, but does not depend on bazelForkOrURLDirName.
+// This is important, as the directory may be added to $PATH, and varying the path for equivalent files may cause unnecessary repository rule cache invalidations.
+// Where a file was downloaded from shouldn't affect cache behaviour of Bazel invocations.
+//
+// The structure of the downloads directory is as follows ([]s indicate variables):
+//
+//	downloads/metadata/[fork-or-url]/bazel-[version-os-etc] is a text file containing a hex sha256 of the contents of the downloaded bazel file.
+//	downloads/sha256/[sha256]/bin/bazel[extension] contains the bazel with a particular sha256.
+func downloadBazelIfNecessary(version string, bazeliskHome string, bazelForkOrURLDirName string, repos *Repositories, config config.Config, downloader DownloadFunc) (string, error) {
 	pathSegment, err := platforms.DetermineBazelFilename(version, false)
 	if err != nil {
 		return "", fmt.Errorf("could not determine path segment to use for Bazel binary: %v", err)
 	}
 
-	destDir := filepath.Join(baseDirectory, pathSegment, "bin")
-	expectedSha256 := strings.ToLower(GetEnvOrConfig("BAZELISK_VERIFY_SHA256"))
-
-	tmpDestFile := "bazel-tmp" + platforms.DetermineExecutableFilenameSuffix()
 	destFile := "bazel" + platforms.DetermineExecutableFilenameSuffix()
 
-	destPath := filepath.Join(destDir, destFile)
-	if _, err := os.Stat(destPath); err == nil {
-		return destPath, nil
+	mappingPath := filepath.Join(bazeliskHome, "downloads", "metadata", bazelForkOrURLDirName, pathSegment)
+	digestFromMappingFile, err := os.ReadFile(mappingPath)
+	if err == nil {
+		pathToBazelInCAS := filepath.Join(bazeliskHome, "downloads", "sha256", string(digestFromMappingFile), "bin", destFile)
+		if _, err := os.Stat(pathToBazelInCAS); err == nil {
+			return pathToBazelInCAS, nil
+		}
 	}
 
+	pathToBazelInCAS, downloadedDigest, err := downloadBazelToCAS(version, bazeliskHome, repos, config, downloader)
+	if err != nil {
+		return "", fmt.Errorf("failed to download bazel: %w", err)
+	}
+
+	expectedSha256 := strings.ToLower(config.Get("BAZELISK_VERIFY_SHA256"))
+	if len(expectedSha256) > 0 {
+		if expectedSha256 != downloadedDigest {
+			return "", fmt.Errorf("%s has sha256=%s but need sha256=%s", pathToBazelInCAS, downloadedDigest, expectedSha256)
+		}
+	}
+
+	if err := atomicWriteFile(mappingPath, []byte(downloadedDigest), 0644); err != nil {
+		return "", fmt.Errorf("failed to write mapping file after downloading bazel: %w", err)
+	}
+
+	return pathToBazelInCAS, nil
+}
+
+func atomicWriteFile(path string, contents []byte, perm os.FileMode) error {
+	if err := os.MkdirAll(filepath.Dir(path), 0755); err != nil {
+		return fmt.Errorf("failed to MkdirAll parent of %s: %w", path, err)
+	}
+	tmpPath := path + ".tmp"
+	if err := os.WriteFile(tmpPath, contents, perm); err != nil {
+		return fmt.Errorf("failed to write file %s: %w", tmpPath, err)
+	}
+	if err := os.Rename(tmpPath, path); err != nil {
+		return fmt.Errorf("failed to rename %s to %s: %w", tmpPath, path, err)
+	}
+	return nil
+}
+
+func downloadBazelToCAS(version string, bazeliskHome string, repos *Repositories, config config.Config, downloader DownloadFunc) (string, string, error) {
+	downloadsDir := filepath.Join(bazeliskHome, "downloads")
+	temporaryDownloadDir := filepath.Join(downloadsDir, "_tmp")
+	casDir := filepath.Join(bazeliskHome, "downloads", "sha256")
+
+	tmpDestFileBytes := make([]byte, 32)
+	if _, err := rand.Read(tmpDestFileBytes); err != nil {
+		return "", "", fmt.Errorf("failed to generate temporary file name: %w", err)
+	}
+	tmpDestFile := fmt.Sprintf("%x", tmpDestFileBytes)
+
 	var tmpDestPath string
-	baseURL := GetEnvOrConfig(BaseURLEnv)
-	formatURL := GetEnvOrConfig(FormatURLEnv)
+	var err error
+	baseURL := config.Get(BaseURLEnv)
+	formatURL := config.Get(FormatURLEnv)
 	if baseURL != "" && formatURL != "" {
-		return "", fmt.Errorf("cannot set %s and %s at once", BaseURLEnv, FormatURLEnv)
+		return "", "", fmt.Errorf("cannot set %s and %s at once", BaseURLEnv, FormatURLEnv)
 	} else if formatURL != "" {
-		tmpDestPath, err = repos.DownloadFromFormatURL(formatURL, version, destDir, tmpDestFile)
+		tmpDestPath, err = repos.DownloadFromFormatURL(config, formatURL, version, temporaryDownloadDir, tmpDestFile)
 	} else if baseURL != "" {
-		tmpDestPath, err = repos.DownloadFromBaseURL(baseURL, version, destDir, tmpDestFile)
+		tmpDestPath, err = repos.DownloadFromBaseURL(baseURL, version, temporaryDownloadDir, tmpDestFile)
 	} else {
-		tmpDestPath, err = downloader(destDir, tmpDestFile)
+		tmpDestPath, err = downloader(temporaryDownloadDir, tmpDestFile)
 	}
 	if err != nil {
-		return "", err
+		return "", "", fmt.Errorf("failed to download bazel: %w", err)
 	}
 
-	if len(expectedSha256) > 0 {
-		f, err := os.Open(tmpDestPath)
-		if err != nil {
-			os.Remove(tmpDestPath)
-			return "", fmt.Errorf("cannot open %s after download: %v", tmpDestPath, err)
-		}
-		defer os.Remove(tmpDestPath)
-		// We cannot defer f.Close() because keeping the handle open when we try to do the
-		// rename later on fails on Windows.
+	f, err := os.Open(tmpDestPath)
+	if err != nil {
+		return "", "", fmt.Errorf("failed to open downloaded bazel to digest it: %w", err)
+	}
 
-		h := sha256.New()
-		if _, err := io.Copy(h, f); err != nil {
-			f.Close()
-			return "", fmt.Errorf("cannot compute sha256 of %s after download: %v", tmpDestPath, err)
-		}
+	h := sha256.New()
+	if _, err := io.Copy(h, f); err != nil {
 		f.Close()
+		return "", "", fmt.Errorf("cannot compute sha256 of %s after download: %v", tmpDestPath, err)
+	}
+	f.Close()
+	actualSha256 := strings.ToLower(fmt.Sprintf("%x", h.Sum(nil)))
 
-		actualSha256 := strings.ToLower(fmt.Sprintf("%x", h.Sum(nil)))
-		if expectedSha256 != actualSha256 {
-			return "", fmt.Errorf("%s has sha256=%s but need sha256=%s", tmpDestPath, actualSha256, expectedSha256)
-		}
+	pathToBazelInCAS := filepath.Join(casDir, actualSha256, "bin", "bazel"+platforms.DetermineExecutableFilenameSuffix())
+	if err := os.MkdirAll(filepath.Dir(pathToBazelInCAS), 0755); err != nil {
+		return "", "", fmt.Errorf("failed to MkdirAll parent of %s: %w", pathToBazelInCAS, err)
 	}
 
-	// Only place the downloaded binary in its final location once we know it is fully downloaded
-	// and valid, to prevent invalid files from ever being executed.
-	if err = os.Rename(tmpDestPath, destPath); err != nil {
-		return "", fmt.Errorf("cannot rename %s to %s: %v", tmpDestPath, destPath, err)
+	tmpPathInCorrectDirectory := pathToBazelInCAS + ".tmp"
+	if err := os.Rename(tmpDestPath, tmpPathInCorrectDirectory); err != nil {
+		return "", "", fmt.Errorf("failed to move %s to %s: %w", tmpDestPath, tmpPathInCorrectDirectory, err)
 	}
-	return destPath, nil
+	if err := os.Rename(tmpPathInCorrectDirectory, pathToBazelInCAS); err != nil {
+		return "", "", fmt.Errorf("failed to move %s to %s: %w", tmpPathInCorrectDirectory, pathToBazelInCAS, err)
+	}
+
+	return pathToBazelInCAS, actualSha256, nil
 }
 
 func copyFile(src, dst string, perm os.FileMode) error {
@@ -524,24 +491,24 @@ func linkLocalBazel(baseDirectory string, bazelPath string) (string, error) {
 	return destinationPath, nil
 }
 
-func maybeDelegateToWrapperFromDir(bazel string, wd string, ignoreEnv bool) string {
-	if !ignoreEnv && GetEnvOrConfig(skipWrapperEnv) != "" {
+func maybeDelegateToWrapperFromDir(bazel string, wd string, config config.Config) string {
+	if config.Get(skipWrapperEnv) != "" {
 		return bazel
 	}
 
-	root := findWorkspaceRoot(wd)
+	root := ws.FindWorkspaceRoot(wd)
 	wrapper := filepath.Join(root, wrapperPath)
 	if stat, err := os.Stat(wrapper); err == nil && !stat.Mode().IsDir() && stat.Mode().Perm()&0111 != 0 {
 		return wrapper
 	}
 
 	if runtime.GOOS == "windows" {
-		powershellWrapper := filepath.Join(root, wrapperPath + ".ps1")
+		powershellWrapper := filepath.Join(root, wrapperPath+".ps1")
 		if stat, err := os.Stat(powershellWrapper); err == nil && !stat.Mode().IsDir() {
 			return powershellWrapper
 		}
 
-		batchWrapper := filepath.Join(root, wrapperPath + ".bat")
+		batchWrapper := filepath.Join(root, wrapperPath+".bat")
 		if stat, err := os.Stat(batchWrapper); err == nil && !stat.Mode().IsDir() {
 			return batchWrapper
 		}
@@ -550,13 +517,13 @@ func maybeDelegateToWrapperFromDir(bazel string, wd string, ignoreEnv bool) stri
 	return bazel
 }
 
-func maybeDelegateToWrapper(bazel string) string {
+func maybeDelegateToWrapper(bazel string, config config.Config) string {
 	wd, err := os.Getwd()
 	if err != nil {
 		return bazel
 	}
 
-	return maybeDelegateToWrapperFromDir(bazel, wd, false)
+	return maybeDelegateToWrapperFromDir(bazel, wd, config)
 }
 
 func prependDirToPathList(cmd *exec.Cmd, dir string) {
@@ -578,8 +545,8 @@ func prependDirToPathList(cmd *exec.Cmd, dir string) {
 	}
 }
 
-func makeBazelCmd(bazel string, args []string, out io.Writer) *exec.Cmd {
-	execPath := maybeDelegateToWrapper(bazel)
+func makeBazelCmd(bazel string, args []string, out io.Writer, config config.Config) *exec.Cmd {
+	execPath := maybeDelegateToWrapper(bazel, config)
 
 	cmd := exec.Command(execPath, args...)
 	cmd.Env = append(os.Environ(), skipWrapperEnv+"=true")
@@ -597,8 +564,8 @@ func makeBazelCmd(bazel string, args []string, out io.Writer) *exec.Cmd {
 	return cmd
 }
 
-func runBazel(bazel string, args []string, out io.Writer) (int, error) {
-	cmd := makeBazelCmd(bazel, args, out)
+func runBazel(bazel string, args []string, out io.Writer, config config.Config) (int, error) {
+	cmd := makeBazelCmd(bazel, args, out, config)
 	err := cmd.Start()
 	if err != nil {
 		return 1, fmt.Errorf("could not start Bazel: %v", err)
@@ -627,14 +594,14 @@ func runBazel(bazel string, args []string, out io.Writer) (int, error) {
 }
 
 // getIncompatibleFlags returns all incompatible flags for the current Bazel command in alphabetical order.
-func getIncompatibleFlags(bazelPath, cmd string) ([]string, error) {
-	var incompatibleFlagsStr = GetEnvOrConfig("BAZELISK_INCOMPATIBLE_FLAGS")
+func getIncompatibleFlags(bazelPath, cmd string, config config.Config) ([]string, error) {
+	var incompatibleFlagsStr = config.Get("BAZELISK_INCOMPATIBLE_FLAGS")
 	if len(incompatibleFlagsStr) > 0 {
 		return strings.Split(incompatibleFlagsStr, ","), nil
 	}
 
 	out := strings.Builder{}
-	if _, err := runBazel(bazelPath, []string{"help", cmd, "--short"}, &out); err != nil {
+	if _, err := runBazel(bazelPath, []string{"help", cmd, "--short"}, &out, config); err != nil {
 		return nil, fmt.Errorf("unable to determine incompatible flags with binary %s: %v", bazelPath, err)
 	}
 
@@ -670,27 +637,27 @@ func insertArgs(baseArgs []string, newArgs []string) []string {
 func parseStartupOptions(baseArgs []string) []string {
 	var result []string
 	var bazelCommands = map[string]bool{
-		"analyze-profile": true,
-		"aquery": true,
-		"build": true,
+		"analyze-profile":    true,
+		"aquery":             true,
+		"build":              true,
 		"canonicalize-flags": true,
-		"clean": true,
-		"coverage": true,
-		"cquery": true,
-		"dump": true,
-		"fetch": true,
-		"help": true,
-		"info": true,
-		"license": true,
-		"mobile-install": true,
-		"mod": true,
-		"print_action": true,
-		"query": true,
-		"run": true,
-		"shutdown": true,
-		"sync": true,
-		"test": true,
-		"version": true,
+		"clean":              true,
+		"coverage":           true,
+		"cquery":             true,
+		"dump":               true,
+		"fetch":              true,
+		"help":               true,
+		"info":               true,
+		"license":            true,
+		"mobile-install":     true,
+		"mod":                true,
+		"print_action":       true,
+		"query":              true,
+		"run":                true,
+		"shutdown":           true,
+		"sync":               true,
+		"test":               true,
+		"version":            true,
 	}
 	// Arguments before a Bazel command are startup options.
 	for _, arg := range baseArgs {
@@ -702,15 +669,15 @@ func parseStartupOptions(baseArgs []string) []string {
 	return result
 }
 
-func shutdownIfNeeded(bazelPath string, startupOptions []string) {
-	bazeliskClean := GetEnvOrConfig("BAZELISK_SHUTDOWN")
+func shutdownIfNeeded(bazelPath string, startupOptions []string, config config.Config) {
+	bazeliskClean := config.Get("BAZELISK_SHUTDOWN")
 	if len(bazeliskClean) == 0 {
 		return
 	}
 
 	args := append(startupOptions, "shutdown")
 	fmt.Printf("bazel %s\n", strings.Join(args, " "))
-	exitCode, err := runBazel(bazelPath, args, nil)
+	exitCode, err := runBazel(bazelPath, args, nil, config)
 	fmt.Printf("\n")
 	if err != nil {
 		log.Fatalf("failed to run bazel shutdown: %v", err)
@@ -721,15 +688,15 @@ func shutdownIfNeeded(bazelPath string, startupOptions []string) {
 	}
 }
 
-func cleanIfNeeded(bazelPath string, startupOptions []string) {
-	bazeliskClean := GetEnvOrConfig("BAZELISK_CLEAN")
+func cleanIfNeeded(bazelPath string, startupOptions []string, config config.Config) {
+	bazeliskClean := config.Get("BAZELISK_CLEAN")
 	if len(bazeliskClean) == 0 {
 		return
 	}
 
 	args := append(startupOptions, "clean", "--expunge")
 	fmt.Printf("bazel %s\n", strings.Join(args, " "))
-	exitCode, err := runBazel(bazelPath, args, nil)
+	exitCode, err := runBazel(bazelPath, args, nil, config)
 	fmt.Printf("\n")
 	if err != nil {
 		log.Fatalf("failed to run clean: %v", err)
@@ -745,17 +712,17 @@ type ParentCommit struct {
 }
 
 type Commit struct {
-	SHA string `json:"sha"`
+	SHA     string         `json:"sha"`
 	PARENTS []ParentCommit `json:"parents"`
 }
 
 type CompareResponse struct {
-	Commits []Commit `json:"commits"`
-	BaseCommit Commit `json:"base_commit"`
-	MergeBaseCommit Commit `json:"merge_base_commit"`
+	Commits         []Commit `json:"commits"`
+	BaseCommit      Commit   `json:"base_commit"`
+	MergeBaseCommit Commit   `json:"merge_base_commit"`
 }
 
-func sendRequest(url string) (*http.Response, error) {
+func sendRequest(url string, config config.Config) (*http.Response, error) {
 	client := &http.Client{}
 
 	req, err := http.NewRequest("GET", url, nil)
@@ -763,7 +730,7 @@ func sendRequest(url string) (*http.Response, error) {
 		return nil, err
 	}
 
-	githubToken := GetEnvOrConfig("BAZELISK_GITHUB_TOKEN")
+	githubToken := config.Get("BAZELISK_GITHUB_TOKEN")
 	if len(githubToken) != 0 {
 		req.Header.Set("Authorization", fmt.Sprintf("token %s", githubToken))
 	}
@@ -771,7 +738,7 @@ func sendRequest(url string) (*http.Response, error) {
 	return client.Do(req)
 }
 
-func getBazelCommitsBetween(goodCommit string, badCommit string) (string, []string, error) {
+func getBazelCommitsBetween(goodCommit string, badCommit string, config config.Config) (string, []string, error) {
 	commitList := make([]string, 0)
 	page := 1
 	perPage := 250 // 250 is the maximum number of commits per page
@@ -779,7 +746,7 @@ func getBazelCommitsBetween(goodCommit string, badCommit string) (string, []stri
 	for {
 		url := fmt.Sprintf("https://api.github.com/repos/bazelbuild/bazel/compare/%s...%s?page=%d&per_page=%d", goodCommit, badCommit, page, perPage)
 
-		response, err := sendRequest(url)
+		response, err := sendRequest(url, config)
 		if err != nil {
 			return goodCommit, nil, fmt.Errorf("Error fetching commit data: %v", err)
 		}
@@ -836,11 +803,11 @@ func getBazelCommitsBetween(goodCommit string, badCommit string) (string, []stri
 	return goodCommit, commitList, nil
 }
 
-func bisect(goodCommit string, badCommit string, args []string, bazeliskHome string, repos *Repositories) {
+func bisect(goodCommit string, badCommit string, args []string, bazeliskHome string, repos *Repositories, config config.Config) {
 
 	// 1. Get the list of commits between goodCommit and badCommit
 	fmt.Printf("\n\n--- Getting the list of commits between %s and %s\n\n", goodCommit, badCommit)
-	goodCommit, commitList, err := getBazelCommitsBetween(goodCommit, badCommit)
+	goodCommit, commitList, err := getBazelCommitsBetween(goodCommit, badCommit, config)
 	if err != nil {
 		log.Fatalf("Failed to get commits: %v", err)
 		os.Exit(1)
@@ -848,7 +815,7 @@ func bisect(goodCommit string, badCommit string, args []string, bazeliskHome str
 
 	// 2. Check if goodCommit is actually good
 	fmt.Printf("\n\n--- Verifying if the given good Bazel commit (%s) is actually good\n\n", goodCommit)
-	bazelExitCode, err := testWithBazelAtCommit(goodCommit, args, bazeliskHome, repos)
+	bazelExitCode, err := testWithBazelAtCommit(goodCommit, args, bazeliskHome, repos, config)
 	if err != nil {
 		log.Fatalf("could not run Bazel: %v", err)
 		os.Exit(1)
@@ -865,8 +832,8 @@ func bisect(goodCommit string, badCommit string, args []string, bazeliskHome str
 	for left < right {
 		mid := (left + right) / 2
 		midCommit := commitList[mid]
-		fmt.Printf("\n\n--- Testing with Bazel built at %s, %d commits remaining...\n\n", midCommit, right -left)
-		bazelExitCode, err := testWithBazelAtCommit(midCommit, args, bazeliskHome, repos)
+		fmt.Printf("\n\n--- Testing with Bazel built at %s, %d commits remaining...\n\n", midCommit, right-left)
+		bazelExitCode, err := testWithBazelAtCommit(midCommit, args, bazeliskHome, repos, config)
 		if err != nil {
 			log.Fatalf("could not run Bazel: %v", err)
 			os.Exit(1)
@@ -892,16 +859,16 @@ func bisect(goodCommit string, badCommit string, args []string, bazeliskHome str
 	os.Exit(0)
 }
 
-func testWithBazelAtCommit(bazelCommit string, args []string, bazeliskHome string, repos *Repositories) (int, error) {
-	bazelPath, err := downloadBazel(bazelCommit, bazeliskHome, repos)
+func testWithBazelAtCommit(bazelCommit string, args []string, bazeliskHome string, repos *Repositories, config config.Config) (int, error) {
+	bazelPath, err := downloadBazel(bazelCommit, bazeliskHome, repos, config)
 	if err != nil {
 		return 1, fmt.Errorf("could not download Bazel: %v", err)
 	}
 	startupOptions := parseStartupOptions(args)
-	shutdownIfNeeded(bazelPath, startupOptions)
-	cleanIfNeeded(bazelPath, startupOptions)
+	shutdownIfNeeded(bazelPath, startupOptions, config)
+	cleanIfNeeded(bazelPath, startupOptions, config)
 	fmt.Printf("bazel %s\n", strings.Join(args, " "))
-	bazelExitCode, err := runBazel(bazelPath, args, nil)
+	bazelExitCode, err := runBazel(bazelPath, args, nil, config)
 	if err != nil {
 		return -1, fmt.Errorf("could not run Bazel: %v", err)
 	}
@@ -909,16 +876,16 @@ func testWithBazelAtCommit(bazelCommit string, args []string, bazeliskHome strin
 }
 
 // migrate will run Bazel with each flag separately and report which ones are failing.
-func migrate(bazelPath string, baseArgs []string, flags []string) {
+func migrate(bazelPath string, baseArgs []string, flags []string, config config.Config) {
 	var startupOptions = parseStartupOptions(baseArgs)
 
 	// 1. Try with all the flags.
 	args := insertArgs(baseArgs, flags)
 	fmt.Printf("\n\n--- Running Bazel with all incompatible flags\n\n")
-	shutdownIfNeeded(bazelPath, startupOptions)
-	cleanIfNeeded(bazelPath, startupOptions)
+	shutdownIfNeeded(bazelPath, startupOptions, config)
+	cleanIfNeeded(bazelPath, startupOptions, config)
 	fmt.Printf("bazel %s\n", strings.Join(args, " "))
-	exitCode, err := runBazel(bazelPath, args, nil)
+	exitCode, err := runBazel(bazelPath, args, nil, config)
 	if err != nil {
 		log.Fatalf("could not run Bazel: %v", err)
 	}
@@ -930,10 +897,10 @@ func migrate(bazelPath string, baseArgs []string, flags []string) {
 	// 2. Try with no flags, as a sanity check.
 	args = baseArgs
 	fmt.Printf("\n\n--- Running Bazel with no incompatible flags\n\n")
-	shutdownIfNeeded(bazelPath, startupOptions)
-	cleanIfNeeded(bazelPath, startupOptions)
+	shutdownIfNeeded(bazelPath, startupOptions, config)
+	cleanIfNeeded(bazelPath, startupOptions, config)
 	fmt.Printf("bazel %s\n", strings.Join(args, " "))
-	exitCode, err = runBazel(bazelPath, args, nil)
+	exitCode, err = runBazel(bazelPath, args, nil, config)
 	if err != nil {
 		log.Fatalf("could not run Bazel: %v", err)
 	}
@@ -948,10 +915,10 @@ func migrate(bazelPath string, baseArgs []string, flags []string) {
 	for _, arg := range flags {
 		args = insertArgs(baseArgs, []string{arg})
 		fmt.Printf("\n\n--- Running Bazel with %s\n\n", arg)
-		shutdownIfNeeded(bazelPath, startupOptions)
-		cleanIfNeeded(bazelPath, startupOptions)
+		shutdownIfNeeded(bazelPath, startupOptions, config)
+		cleanIfNeeded(bazelPath, startupOptions, config)
 		fmt.Printf("bazel %s\n", strings.Join(args, " "))
-		exitCode, err = runBazel(bazelPath, args, nil)
+		exitCode, err = runBazel(bazelPath, args, nil, config)
 		if err != nil {
 			log.Fatalf("could not run Bazel: %v", err)
 		}
